from typing import Dict, Tuple

class Pipeline:

    _dataFrameLoaders: Dict[str, callable] = dict()
    _transformations: Dict[str, callable] = dict()
    _sources: Dict[str, Tuple[callable]] = dict()
    _services: Dict[str, Tuple[object]] = dict()

    def addDataFrameLoader(self, fun: callable, services: Tuple[object]):
        functionName = fun.__name__

        if functionName not in self._dataFrameLoaders:
            self._dataFrameLoaders[functionName] = fun
            self._services[functionName] = services

    def addTransformation(self, fun: callable, sources: Tuple[callable], services: Tuple[object]):
        functionName = fun.__name__

        if functionName not in self._transformations:
            self._transformations[functionName] = fun
            self._sources[functionName] = sources
            self._services[functionName] = services

    def getSource(self, functionName) -> Tuple[callable]:
        return self._sources[functionName]

    def getServices(self, functionName) -> Tuple[object]:
        return self._services[functionName]

    def getTransformations(self) -> Dict[str, callable]:
        return self._transformations

    def getDataFrameLoaders(self) -> Dict[str, callable]:
        return self._dataFrameLoaders



# pylint: disable = invalid-name
import os
import sys
from pathlib import Path
from typing import Tuple
from databricksbundle.pipeline.function.ServicesResolver import ServicesResolver
from databricksbundle.pipeline.Pipeline import Pipeline
from pyfonybundles.appContainerInit import initAppContainer
from databricksbundle.pipeline.decorator.static_init import static_init
from databricksbundle.pipeline.decorator.argsChecker import checkArgs
from databricksbundle.pipeline.pipelineInvoker import invokePipeline

@static_init
class PipelineDecorator:

    _servicesResolver: ServicesResolver
    pipeline: Pipeline

    @classmethod
    def static_init(cls):
        container = initAppContainer(os.environ['APP_ENV'])

        cls._servicesResolver = container.get(ServicesResolver)
        cls.pipeline = Pipeline()

class dataFrameLoader(PipelineDecorator):

    def __init__(self, *args, **kwargs): # pylint: disable = unused-argument
        checkArgs(args)

    def __call__(self, fun, *args, **kwargs):
        pipelinePath = Path(sys.argv[0])
        services = self._servicesResolver.resolve(fun, 0, pipelinePath) # pylint: disable = no-member

        PipelineDecorator.pipeline.addDataFrameLoader(fun, services)
        return fun

class transformation(PipelineDecorator):

    def __init__(self, *args: Tuple[callable], **kwargs): # pylint: disable = unused-argument
        self._sources = args

    def __call__(self, fun, *args, **kwargs):
        startIndex = len(self._sources)
        pipelinePath = Path(sys.argv[0])
        services = self._servicesResolver.resolve(fun, startIndex, pipelinePath) # pylint: disable = no-member
        PipelineDecorator.pipeline.addTransformation(fun, self._sources, services)
        return fun

class dataFrameSaver(PipelineDecorator):

    def __init__(self, *args: Tuple[callable]):
        self._sources = args

    def __call__(self, fun, *args, **kwargs):
        dataFramesByName = invokePipeline(PipelineDecorator.pipeline)

        def transformSource(source: callable):
            return dataFramesByName[source.__name__ + '_df']

        pipelinePath = Path(sys.argv[0])
        dataframesToUse = tuple(map(transformSource, self._sources))
        services = self._servicesResolver.resolve(fun, 1, pipelinePath) # pylint: disable = no-member

        fun(*(dataframesToUse + services))

        return fun



from typing import Dict
from pyspark.sql.dataframe import DataFrame
from databricksbundle.pipeline.Pipeline import Pipeline

def invokePipeline(pipeline: Pipeline) -> Dict[str, DataFrame]:
    dataFramesByName: Dict[str, DataFrame] = dict()

    for dataFrameLoaderName, dataFrameLoader in pipeline.getDataFrameLoaders().items():
        services = pipeline.getServices(dataFrameLoaderName)
        dataFramesByName[dataFrameLoaderName + '_df'] = dataFrameLoader(*services)

    for transformationName, transformation in pipeline.getTransformations().items():
        def transformSource(source: callable):
            return dataFramesByName[source.__name__ + '_df']

        dataframesToUse = tuple(map(transformSource, pipeline.getSource(transformationName)))
        services = pipeline.getServices(transformationName)

        df = transformation(*(dataframesToUse + services))
        dataFramesByName[transformationName + '_df'] = df

    return dataFramesByName



# pylint: disable = unused-import
import sys
from databricksbundle.detector import isDatabricks

if isDatabricks():
    from databricksbundle.pipeline.decorator.environment.databricks import dataFrameLoader, transformation, dataFrameSaver
elif 'unittest' in sys.modules:
    from databricksbundle.pipeline.decorator.environment.unittest import dataFrameLoader, transformation, dataFrameSaver
else:
    from databricksbundle.pipeline.decorator.environment.pyscript import dataFrameLoader, transformation, dataFrameSaver



# pylint: disable = invalid-name
from databricksbundle.pipeline.decorator.argsChecker import checkArgs

class dataFrameLoader:

    def __init__(self, *args, **kwargs):  # pylint: disable = unused-argument
        checkArgs(args)

    def __call__(self, fun, *args, **kwargs):
        return fun

class transformation:

    def __init__(self, *args, **kwargs):
        pass

    def __call__(self, fun, *args, **kwargs):
        return fun

class dataFrameSaver:

    def __init__(self, *args):
        pass

    def __call__(self, fun, *args, **kwargs):
        return fun



# pylint: disable = invalid-name
import os
from pathlib import Path
from typing import Tuple
from databricksbundle.display import display
from databricksbundle.pipeline.function.ServicesResolver import ServicesResolver
from databricksbundle.pipeline.decorator.static_init import static_init
from databricksbundle.pipeline.decorator.argsChecker import checkArgs
from injecta.dtype.classLoader import loadClass
from databricksbundle.notebook.helpers import getNotebookPath

@static_init
class PipelineDecorator:

    _servicesResolver: ServicesResolver

    @classmethod
    def static_init(cls):
        if 'CONTAINER_INIT_FUNCTION' not in os.environ:
            raise Exception('CONTAINER_INIT_FUNCTION environment variable must be set on cluster')

        containerInitFunctionPath = os.environ['CONTAINER_INIT_FUNCTION']
        containerInitModuleName = containerInitFunctionPath[0:containerInitFunctionPath.rfind('.')]
        containerInitFunctionName = containerInitFunctionPath[containerInitFunctionPath.rfind('.') + 1:]

        containerInitFunction = loadClass(containerInitModuleName, containerInitFunctionName)
        container = containerInitFunction(os.environ['APP_ENV'])

        cls._servicesResolver = container.get(ServicesResolver)

class dataFrameLoader(PipelineDecorator):

    def __init__(self, *args, **kwargs):
        self._displayEnabled = kwargs.get('display', False)
        checkArgs(args)

    def __call__(self, fun, *args, **kwargs):
        pipelinePath = Path(getNotebookPath())
        services = self._servicesResolver.resolve(fun, 0, pipelinePath) # pylint: disable = no-member

        g = fun.__globals__
        df = fun(*services)
        g[fun.__name__ + '_df'] = df

        if self._displayEnabled:
            display(df)

        return fun

class transformation(PipelineDecorator):

    def __init__(self, *args: Tuple[callable], **kwargs):
        self._sources = args
        self._displayEnabled = kwargs.get('display', False)

    def __call__(self, fun, *args, **kwargs):
        g = fun.__globals__

        def transformSource(source: callable):
            return g[source.__name__ + '_df']

        pipelinePath = Path(getNotebookPath())
        dataframesToUse = tuple(map(transformSource, self._sources))
        startIndex = len(self._sources)
        services = self._servicesResolver.resolve(fun, startIndex, pipelinePath) # pylint: disable = no-member

        df = fun(*(dataframesToUse + services))
        g[fun.__name__ + '_df'] = df

        if self._displayEnabled:
            display(df)

        return fun

class dataFrameSaver(PipelineDecorator):

    def __init__(self, *args: Tuple[callable]):
        self._sources = args

    def __call__(self, fun, *args, **kwargs):
        g = fun.__globals__

        def transformSource(source: callable):
            return g[source.__name__ + '_df']

        pipelinePath = Path(getNotebookPath())
        dataframesToUse = tuple(map(transformSource, self._sources))
        services = self._servicesResolver.resolve(fun, 1, pipelinePath) # pylint: disable = no-member

        fun(*(dataframesToUse + services))

        return fun



from types import FunctionType

def checkArgs(a):
    # called as @dataFrameLoader (without brackets and any arguments)
    if a and isinstance(a[0], FunctionType):
        raise Exception('Use @dataFrameLoader() instead of @dataFrameLoader please')
